#include <functional>
#include <iostream>
#include <numeric>
#include <random>
#include <valarray>

// 不失一般性, k = 1

struct lattice {
  int dimension;                    // 系统维度
  int lattice_size;                 // 格子的边长
  int total_size;                   // 格子总大小, 即面积或者体积
  std::vector<double> lattice_data; // 系统的自旋, 每个点为+0.5或者-0.5
  double mean_mu_z;                 // 格子的z方向平均自旋
  std ::mt19937 engine;

  // 通过维度， 系统边长， 以及随机数种子来初始化系统
  lattice(int dimension, int lattice_size, int seed)
      : dimension(dimension), lattice_size(lattice_size),
        total_size(std::pow(lattice_size, dimension)) {
    lattice_data.resize(total_size);
    engine = std::mt19937(seed);
    auto dist = std::uniform_int_distribution<int>(0, 1);
    std::generate(lattice_data.begin(), lattice_data.end(),
                  [&]() { return -0.5; });
    double sum = std::accumulate(lattice_data.begin(), lattice_data.end(), 0.);
    // std::cout << sum << "SUM\n";
    mean_mu_z = sum / total_size;
  }

  // 根据坐标获取这个格点的自旋
  double &get(const std::vector<int> &position) {
    int offset = 0;
    for (const auto &i : position) {
      offset = offset * lattice_size + i;
    }
    return lattice_data[offset];
  }

  // 给定J，h，T进行一部metropolis采样，并计算mean_mu_z
  void hop(double J, double h, double T) {
    auto dist_int = std::uniform_int_distribution<int>(0, lattice_size - 1);
    auto position = std::vector<int>(dimension);
    std::generate(position.begin(), position.end(),
                  [&]() { return dist_int(engine); });
    double delta_E = 0;
    auto current = get(position);
    bool up_to_down = current > 0;
    delta_E += 2 * (-h) * (-2 * current);
    for (int i = 0; i < dimension; i++) {
      for (int j = -1; j <= 1; j += 2) {
        auto neibor = position;
        neibor[i] += j;
        if (neibor[i] != -1 && neibor[i] != lattice_size) {
          delta_E += 2 * (-J) * (-current * get(neibor));
        }
      }
    }
    auto dist_real = std::uniform_real_distribution<double>(0, 1);
    // std::cout << delta_E << ' ' << std::exp(-delta_E / T) << ' ';
    if (dist_real(engine) < std::exp(-delta_E / T)) {
      // std::cout << "h\n";
      if (up_to_down) {
        get(position) = -0.5;
        mean_mu_z -= 1. / total_size;
      } else {
        get(position) = +0.5;
        mean_mu_z += 1. / total_size;
      }
    } else {
      // std::cout << "n\n";
    }
  }
};

// 返回mu z均值的期望和校准差
auto calc_mu_z(int dimension, int lattice_size, int time, int length, double J,
               double h, double T) {
  double m = 0, m2 = 0;
  for (int t = 0; t < time; t++) {
    auto l = lattice(dimension, lattice_size, std::random_device()());
    for (int i = 0; i < length; i++) {
      l.hop(J, h, T);
      // l.show();
      m += l.mean_mu_z;
      m2 += l.mean_mu_z * l.mean_mu_z;
    }
#if 0
    for (auto i : l.lattice_data) {
      std::cout << (i > 0) << ' ';
    }
    std::cout << '\n';
    std::cout << l.mean_mu_z;
    std::cout << "\n\n";
#endif
  }
  int number = length * time;
  auto mean_m = m / number;
  auto mean_m2 = m2 / number;
  std::cout << mean_m << "\t" << std::pow(-mean_m * mean_m + mean_m2, 0.5)
            << "\n";
}
